// This is a roxen module. Copyright  2001, Roxen IS.

constant cvs_version =
  "$Id: auth_httpbasic.pike,v 1.8 2001/09/03 18:40:48 nilsson Exp $";
inherit AuthModule;
inherit "module";

constant name = "basic";

//<locale-token project="mod_auth_httpbasic">_</locale-token>
#define _(X,Y)	_DEF_LOCALE("mod_auth_httpbasic",X,Y)

#include <module.h>

LocaleString module_name =
  _(1,"Authentication: Password");

LocaleString module_doc =
  _(2,"Authenticate users using basic username/password authentication.");

void create() {
  defvar("401-files", ({ "401.inc" }),
         _(3, "Authentication required message override files"),
         TYPE_STRING_LIST|VAR_PUBLIC,
         _(4,    "If no file matches the list, all directories above the"
                 " wanted file are searched for one of the files in this list.")
   );

  defvar("401-message", #"<html><head>
<title>401 - Authentication failed</title>
</head>
<body alink=\"#000000\" bgcolor=\"#ffffff\" bottommargin=\"0\" leftmargin=\"0\" link=\"#ce5c00\" marginheight=\"2\" marginwidth=\"0\" rightmargin=\"0\" text=\"#333333\" topmargin=\"2\" vlink=\"#ce5c00\">

<center><h1>Authentication failed</h1>
</center>
<if nserious=''>I don't think I can do that, Dave... </if>
</body>
</html>
",
	 _(5, "Authentication failed message"),
	 TYPE_TEXT_FIELD|VAR_PUBLIC,
	 _(6, "What to return when authentication fails."));
}

static array(string) parse_auth_header( mixed header )
{
  array(string) res;
  array(string) handle_header( string header ) 
  {
    string a, b;
    if( sscanf( header, "%[^ ] %s", a, b ) == 2 )
      switch( a )
      {
	case "Basic":
	case "basic":
	  b = MIME.decode_base64( b );
	  if( sscanf( b, "%[^:]:%s", a, b ) == 2 )
	    return ({ a, b });
      }
  };
  if( arrayp( header ) )
  {
    foreach( header, header )
      if( (res = handle_header( header )) && res[0] )
	return res;
  }
  else
    return handle_header( header );
  return ({ 0,0 });
}


static User low_authenticate( RequestID id,
			      string user, string password,
			      UserDB db)
{
  if( User u = db->find_user( user, id ) )
    if( u->password_authenticate( password ) )
      return u;
}

User authenticate( RequestID id, UserDB db )
//! Try to authenticate the request with users from the specified user
//! database. If no @[db] is specified, all datbases in the current
//! configuration are searched in order, then the configuration user
//! database.
//!
//! The return value is the autenticated user.
{
  string password = id->misc->password;
  string user     = id->misc->user;

  mixed header;
  
  if( !user )
    if( header = id->request_headers[ "authorization" ] )
      [user,password] = parse_auth_header( header ) || ({ 0,0 });
    else if( id->realauth )
      sscanf( id->realauth, "%[^:]:%s", user, password );
    else
      return 0; // Not very likely to work...
  
  if( !user || !password )
    return 0;
  
  if( !db )
  {
    User res;
    foreach( id->conf->user_databases(), UserDB db )
      if( res = low_authenticate( id, user, password, db ) )
	return res;
    return 0;
  }
  return low_authenticate( id, user, password, db );
}


mapping authenticate_throw( RequestID id, string realm, UserDB db )
//! Returns a reply mapping, similar to @[Roxen.http_rxml_reply] with
//! friends. If no @[db] is specified,  all datbases in the current
//! configuration are searched in order, then the configuration user
//! database.
{
  array files = query("401-files");
  array errf;
  if( files && sizeof(files))
  {
    id->misc->authenticate_throw=1;
    errf = VFS.find_above_read( id->not_query, files[0], id, "401", 1);
    m_delete(id->misc,"authenticate_throw");
  }
  return Roxen.http_auth_required( realm,
    errf ? errf[1] : query("401-message"), id );
}
